#from numpy import *
import pylab
from scipy.interpolate import splrep,splev
from numpy import linspace,meshgrid,diag,ones,sign,kron,eye,zeros
from numpy.linalg import inv,norm
from scipy.linalg import block_diag
from scipy import sparse
from sys import exit

class parameters_base:
    """In parameters all data concerning a specific grid and paramters are stored. This also includes stencils, but not solutions."""
    def __init__(self, n_mu, n_z, kappa = 0.385, lmbda=0.0385, xi=1.):
        self.kappa=kappa
        self.lmbda=lmbda
        self.xi = xi

        self.n_mu=n_mu
        self.n_z =n_z
        self.array_mu = linspace(-1.,1.,n_mu+2); self.array_mu=self.array_mu[1:-1]
        self.array_z  = linspace(0.,1.,n_z+2); self.array_z=self.array_z[1:-1]
        self.grid_z,self.grid_mu = meshgrid(self.array_z,self.array_mu)
        self.sel = 0.5*(1-sign(self.array_mu))

        self.matrix = sparse.csc_matrix(self.get_z_matrix()) + sparse.csc_matrix(self.get_mu_matrix())

        self.u0=zeros(self.n_mu*self.n_z)
        self.b=zeros(self.n_mu*self.n_z)
        self.rhs=zeros(self.n_mu*self.n_z)
        self.initial_method=None

    def import_boundary(self,b):
        """Takes the vector b, which describes the Dirichlet inflow boundary conditions and generates the right hand side of the linear equation A·x=rhs accordingly.
        Warning: this function only works if the first order upwind scheme is used at the boundary."""
        self.b = -abs(self.array_mu)*(self.n_z+1)*b
        matrix=zeros((self.n_mu,self.n_z))
        matrix[:,0]=diag(self.sel).dot(self.b)
        matrix[:,-1]=diag(1-self.sel).dot(self.b)
        self.rhs = to_column(matrix,self)

    def get_diff_matrix(self):
        """Returns the mu diffusion centered finite differences stencil."""
        if not hasattr(self,'matrix_cd_mu_neu'):
            mu_mhalf = self.array_mu - 0.5*(self.array_mu[1]-self.array_mu[0])
            mu_phalf = self.array_mu + 0.5*(self.array_mu[1]-self.array_mu[0])

            matrix_cd_mu_neu = diag(1.0-mu_phalf[:-1]**2,1) + diag(1.0-mu_mhalf[1:]**2,-1) - diag(2.0-mu_phalf**2-mu_mhalf**2)
            matrix_cd_mu_neu[0,0]   =-1.0+mu_phalf[0]**2
            matrix_cd_mu_neu[-1,-1] =-1.0+mu_mhalf[-1]**2
            matrix_cd_mu_neu *= (self.n_mu+1)**2/4.

            self.matrix_cd_mu_neu = sparse.csc_matrix(matrix_cd_mu_neu)
        return self.matrix_cd_mu_neu

    def prepare_mu_adv_matrix(self):
        """Returns the finite differences matrix to the mu advection."""
        matrix = diag(ones(self.n_mu-1),1)-diag(ones(self.n_mu-1),-1)
        matrix *= self.n_mu /4.
        matrix = diag(1-self.array_mu**2).dot(matrix)
        return sparse.csr_matrix(matrix)

    def get_mu_matrix_1d(self):
        """Returns the finite difference matrix for all mu processes. It is a build from the advective and diffusive components."""
        return - self.kappa * self.prepare_mu_adv_matrix()/self.xi + self.lmbda* self.get_diff_matrix()/self.xi**3

    def get_mu_matrix(self):
        return sparse.kron(self.get_mu_matrix_1d(),sparse.eye(self.n_z))

    ## The are the placeholders for the z advection finite difference matrices, which are defined in child-classes. The advection direction depends on the sign of mu, hence there are separate up and down matrices.
    def z_stencil_up(self):
        print('Error: this class is not complete. Use one of the child classes.')
        exit(1)
    def z_stencil_down(self):
        print('Error: this class is not complete. Use one of the child classes.')
        exit(1)

    def get_z_matrix(self):
        """The finite differences matrix for the z component is put together with the direction of advection in mind."""
        matrix_up_z = self.z_stencil_up()
        matrix_do_z = self.z_stencil_down()
        matrix_sel_up = sparse.csc_matrix(diag(self.array_mu*self.sel))
        matrix_sel_do = sparse.csc_matrix(diag(self.array_mu*(1.0-self.sel)))

        out = sparse.kron(matrix_sel_up,matrix_up_z) + sparse.kron(matrix_sel_do,matrix_do_z);
        return out;

    def get_z_matrix_inv(self,S=0.1):
        """Compute the inverse of the z-matrix. This can be improved due to the fact, that it is a block diagonal matrix. Hence it is the diagonal of the inverse blocks. """
        lst=[]
        for i,mu in enumerate(self.array_mu):
            if mu<0.0:
                matrix_up_z      = mu*self.z_stencil_up();
                matrix_up_z_inv  = inv(eye(self.n_mu)-S*matrix_up_z)
                lst.append(matrix_up_z_inv)
            else:
                matrix_do_z      = mu*self.z_stencil_down()
                matrix_do_z_inv  = inv(eye(self.n_mu)-S*matrix_do_z)
                lst.append(matrix_do_z_inv)
        return sparse.block_diag(lst)

    def get_Ainv(self,S=0.1):
        """The inverse of A is computed as soon as necessary, and then cached for future use."""
        if not hasattr(self,'Ainv'):
            self.Ainv = sparse.csc_matrix(sparse.kron(sparse.linalg.inv(sparse.csc_matrix(eye(self.n_mu)-S*self.get_mu_matrix_1d())),sparse.eye(self.n_z)))
        return self.Ainv

    def get_linop(self):
        """The linear operator of the inverse ADI is generated and consequently cached as well."""
        if not hasattr(self,'ADIinv'):
            def linop(x):
                return self.get_Ainv().dot(self.get_z_matrix_inv().dot(x))
            self.ADIinv = sparse.linalg.LinearOperator((self.n_mu*self.n_z,self.n_mu*self.n_z),linop,dtype=float)
        return self.ADIinv

class parameters(parameters_base):
    """The first order advection upwind stencils depend on the direction of advection."""
    def z_stencil_up(self):
        return sparse.csc_matrix((self.n_z+1)*(diag(ones(self.n_z)) - diag(ones(self.n_z-1),-1)))
    def z_stencil_down(self):
        return sparse.csc_matrix((self.n_z+1)*(diag(ones(self.n_z-1),1)-diag(ones(self.n_z))))

class parameters_high(parameters_base):
    """This is a variant of the parameters class, where second order stencils are used for the z-advection."""
    def z_stencil_up(self):
        matrix = sparse.csc_matrix((self.n_z+1)/2.*(3*diag(ones(self.n_z)) - 4*diag(ones(self.n_z-1),-1) + diag(ones(self.n_z-2),-2)))
        matrix[0,0]=(self.n_z+1)
        matrix[1,0]=-(self.n_z+1)
        matrix[1,1]=(self.n_z+1)
        return matrix
    def z_stencil_down(self):
        matrix = sparse.csc_matrix((self.n_z+1)/2.*(- diag(ones(self.n_z-2),2) + 4*diag(ones(self.n_z-1),1)-3*diag(ones(self.n_z))))
        matrix[-1,-1]=-(self.n_z+1)
        matrix[-2,-1]=+(self.n_z+1)
        matrix[-2,-2]=-(self.n_z+1)
        return matrix

def to_column(inpt,p):
    """Tronsform a matrix to column major form."""
    return inpt.reshape(p.n_z*p.n_mu)
def from_column(inpt,p):
    """Transform the matrix back from the column major format."""
    return inpt.reshape((p.n_mu,p.n_z))


def maxnorm(x):
    """The maximum norm of the vector/matrix x."""
    return (abs(x.flatten())).max()

def norm2(x):
    """The L2 norm of the vector/matrix x."""
    return (norm(x.flatten(),ord=2))

def normL2(x):
    """The L2 integral norm."""
    print('normL2 not implemented yet.')
    exit(1)

def cost_exp(i_n,i_o,n_n,n_o):
    """Depending on grid size and necessary iterations, a relative computational expense is determined."""
    i_n = float(i_n)
    i_o = float(i_o)
    n_n = float(n_n)
    n_o = float(n_o)

    out = log((i_n * n_n**2)/(i_o *n_o**2))
    out /= log(n_n/n_o)
    if out == 0.: return 1.
    return 0.5 * out

