from basic import *
import json
import os.path
from os import rename
from numpy import reshape,array
from sys import exit
import gc
import base64
import zlib
import datetime

class data(object):
    """Whereas in the parameter classes all (initial) information on a gird is stored, the data class stores the results. It may store arbitrary many results and therefore also keeps a record of all stored key data in the class member info."""
    def __init__(self):
        self.clear()

    def clear(self):
        """Clears the data and info members."""
        self.info={"solver": [], 'n_mu': [], 'n_z': [], 'xi': [], 'kappa': [], 'lambda': [], 'boundary': [], 'parameters': [], 'state': [], 'flags': []};
        if hasattr(self,'data'):
            del self.data;
            gc.collect()
        self.data=[];

    def append_solver(self,solver,save_grid=False,flags=[]):
        """Enters the available results of a given solver into records. If toggled true, also the resulting grid itself is stored."""
        if solver.state==0: return # do not add grid data, that has not been finished yet.
        local_out = {"solver": str(solver.name), "n_mu": solver.parameters.n_mu, "n_z": solver.parameters.n_z, "iterations": solver.iterations, "tolerance": solver.tolerance, "xi": solver.parameters.xi, "kappa": solver.parameters.kappa, "lambda": solver.parameters.lmbda, "boundary": solver.parameters.initial_method, "parameters": str(solver.parameters.__class__.__name__), "state": solver.state, 'flags': flags };
        if solver.error != None: local_out["error"]=solver.error
        if hasattr(solver,'expense'): local_out["cost"]=solver.expense
        if solver.time>0.: local_out["time"]=solver.time
        local_out['finish_time'] = datetime.datetime.now().isoformat()
        if solver.state==1 and save_grid:
            local_out["mu"]=self.array_encode(solver.parameters.array_mu);
            local_out["z"]=self.array_encode(solver.parameters.array_z);
            local_out["data"]=self.array_encode(solver.result);

        self.data.append(local_out);
        self.add_info(local_out)

    def append(self,element):
        """Import a result element as given within. The source may be another instance of the data class."""
        if set(["solver","n_mu","n_z","xi","kappa","lambda","boundary"]).issubset(element) and element not in self.data:
            ### Legacy section -- in early versions the tolerance value was not saved.
            if 'flags' not in element:
                element['flags']=[]
            if element["tolerance"] == 1e-8 and 'reference' not in element['flags']:
                element['flags'].append('reference')
            #if "tolerance" not in element:
                #if element["solver"] == "reference":
                    #element["tolerance"] = 1e-8;
                #else:
                    #element["tolerance"] = 1e-5;
            ## End of legacy section

            self.data.append(element)
            self.add_info(element)

    def size(self):
        return len(self.data)

    def get_state(self,xi,boundary,n_mu,n_z,parameter,solver=None,tolerance=None,flags=[],kappa=None,lmbda=None):
        """Checks whether a specific element is in the storage and if true, it's state is returned. If the element cannot be found state 0 - not computed - is returned."""
        for element in self.data:
            if element['xi'] != xi: continue
            if element['n_mu'] != n_mu: continue
            if element['n_z'] != n_z: continue
            if solver != None and element['solver'] != str(solver.__name__): continue
            if element['parameters'] != str(parameter.__name__): continue
            if tolerance != None and element['tolerance'] != tolerance: continue
            if flags != [] and set(flags).isdisjoint(set(element['flags'])): continue
            if kappa != None and element['kappa'] != kappa: continue
            if lmbda != None and element['lambda'] != lmbda: continue
            return element['state']
        return 0 ## uncomputed/non found

    def import_data(self,data=None):
        """Incorporate all the data from another data instance."""
        if not data: return
        for element in data.data:
            self.append(element);

    def add_info(self,info):
        """Updates the general information info entries."""
        def add_unique(arr, val):
            if val not in arr: arr.append(val);
        for key,val in self.info.items():
            if key=='flags':
                for flag in info['flags']: add_unique(val,flag)
                continue
            add_unique(val,info[key])

    ## Encode and decode array like the grid to an from a more compact compressed representation.
    def array_encode(self,arr):
        tmp = arr
        tmp = " ".join(map(str,reshape(tmp,(1,-1)).flatten()))
        tmp = bytes(tmp,'UTF-8')
        tmp = zlib.compress(tmp,9)
        tmp =  base64.b64encode(tmp)
        tmp =tmp.decode('UTF-8')
        return tmp
    def array_decode(self,blob):
        tmp = base64.b64decode(blob)
        tmp = zlib.decompress(tmp)
        tmp = tmp.decode('UTF-8')
        #tmp = str(tmp)
        #import pdb; pdb.set_trace()
        tmp = tmp.split(" ")
        tmp = [ float(x) for x in tmp ]
        #import pdb; pdb.set_trace()
        return tmp

    def get_grid(self,element):
        """Return the grid with mu and z array in a geometric usefull representation."""
        data,mu,z=[],[],[];
        #import pdb; pdb.set_trace()
        if "data" in element: data=reshape(self.array_decode(element['data']),(element["n_mu"],element["n_z"]));
        if "mu" in element: mu=self.array_decode(element['mu'])
        if "z" in element: z=self.array_decode(element['z'])
        return data,mu,z;

    def get_element_grid(self,pos=0):
        """Wrapper around get_grid that let's one obtain the grid at a specific position in the storage."""
        if pos<0 or pos>len(self.data):
            print('Error: no element at position',pos,'in storage.')
            exit(1)
        return self.get_grid(self.data[pos])

    def sorter(self,var='n_z'):
        """Sort elements by certain values. E.g. the grid size."""
        self.data = sorted(self.data, key=lambda k: k[var]);

    def import_json(self,path=None):
        """Import a records json file."""
        if path == None:
            print("Warning: There is no valid import path. Doing nothing.")
            return
        if  not os.path.isfile(path):
            print('The file',path,'does not exist.')
            exit(1)
            return

        with open(path) as input:
            try: tmp = json.load(input);
            except:
                print('The file',path,'does not have a valid json structure.')
                exit(1)
            for element in tmp["data"]: # only include new elements
                self.append(element);

        self.sorter();

    def export_json(self,path="sim_gmres_out"):
        """Write the records to a json file. At first a temporary file is written, which will be rename after closing the it."""
        if os.path.splitext(path)[1] != "json": path+='.json';
        with open(path+'.tmp','w') as output:
            try: json.dump({"info": self.info, "data": self.data}, output, sort_keys=True, indent=4, separators=(',', ': '));
            except:
                print('Writing json to file',path,'failed. Exiting.')
                exit(1)
        rename(path+'.tmp',path)

    def export_csv(self,keys=[],separator=' '):
        """Output the data to a csv-style list."""
        outstring=''
        for element in self.data:
            for key in keys:
                outstring+=str(element[key])
                outstring+=separator
            outstring+="\n"
        file = open(self.path+'.dat','w')
        file.write(outstring)
        file.close()

