#!/usr/bin/env python3

import json
import sys
import os.path
import matplotlib
matplotlib.use('Agg')
from matplotlib import ticker
from matplotlib.pyplot import *
#import datetime

from numpy import mean,arange #,std,reshape
from mpl_toolkits.axes_grid1 import make_axes_locatable
#import collections
#from scipy.optimize import curve_fit
#import pdb

#from solvers import *
from data import *
from data_evaluation import *

name_color = {'GMRES': 'orange', 'GMRES_right_mu_prec': 'r', 'Richardson': 'k', 'Richardson_right_prec_matrixfree': 'b', 'GMRES_right_prec_matrixfree': 'g'}
bc_marker = {'sin': '^', 'half sin': '<', 'rect': 's', 'sign': 'v', ' forward beam': '>', ' broad beam': 'h', 'gaussian beam': 'o'}
#bc_names = {'initial_sin': '$sin(\frac{\pi}{2}\mu)$', 'initial_sin_sign': 'for $\mu>0: sin(\frac{\pi}{2}\mu)$, else: $0$', 'initial_rect': '$1$', 'initial_sign': 'for $\mu<0: 1$, else: $0$', 'initial_beam': 'for $\mu\geq 0.95: 1$ else $0$', 'initial_beam2': 'for $0.25\leq\mu\leq 0.75: 1$ else $0$', 'initial_beam3': 'for $\mu\geq 0: e^{-100(\mu-0.5)^2}$ else $0$'}
variable_name = {'cost': 'Relative computational expense','error': 'Computational error','iterations': 'Iterations','time': 'Runtime'}
#solver_list={'no': 'GMRES', 'lAinv': 'GMRES_right_mu_prec', 'rADI': 'Richardson_right_prec_matrixfree', 'lADI': 'GMRES_right_prec_matrixfree'}
label_sort=['GMRES', 'GMRES_right_mu_prec', 'Richardson_right_prec_matrixfree', 'GMRES_right_prec_matrixfree']
label_name = {'GMRES': 'GMRES', 'GMRES_right_mu_prec': '$\mu$-preconditioned GMRES', 'GMRES_right_prec_matrixfree': 'ADI preconditioned GMRES', 'Richardson': 'Richardson iteration', 'Richardson_right_prec_matrixfree': 'ADI preconditioned Richardson'}
param_style= {'parameters': '-', 'parameters_high': '-.'}


def print_vals(vals=[]):
    if len(vals)==1:
        return str(vals[0])
    output='\{'
    for val in vals[:-1]:
        output+=str(val)+","
    output+=str(vals[-1])+'\}'
    return output

def averaged_function_of(evaluation,variable='xi',value='iterations',parameters=['parameters', 'parameters_high'],name_color=name_color,param_style=param_style):
    label_vals=[]
    for bound in evaluation.data.info['boundary']:
        for params in parameters:
            for solv in label_sort:
                label=label_name[solv]
                evaluation.reduce(boundaries=[bound],solvers=[solv],parameters=[params]);
                x,y=evaluation.get_variable_data(variable=variable,state=1,value=value,function = lambda x: mean(x));
                #import pdb; pdb.set_trace()
                if solv in name_color: nc = name_color[solv]
                elif params in name_color: nc=name_color[params]
                else: nc='k'
                if solv in param_style: parast=param_style[solv]
                elif params in param_style: parast=param_style[params]
                else: parast='-'
                plot(x,y,color=nc,label=label,marker=bc_marker[bound],linestyle=parast);
                if label not in label_vals: label_vals.append(label)
    return label_vals

def plot_xis(evaluation,out_path,value='iterations'):
    fig = figure()

    label_vals = averaged_function_of(evaluation,variable='xi',value=value)

    ylim([10,30000])

    x_min=1e10; x_max=-1e10;
    for line in gca().lines:
        lin=line.get_xdata()
        if len(lin)<1: break
        x_min=min(x_min,lin[0])
        x_max=max(x_max,lin[-1])
    dist=(x_max-x_min)/48.;
    xlim([x_min-dist,x_max+dist])

    yscale('log')
    xlabel('Normalized velocity $\\xi$')
    ylabel(variable_name[value])

    leg = legend(label_vals,loc='best', prop={'size':10},markerscale=0.)
    leg.get_frame().set_alpha(0.5)
    draw()

    out_path+= '_xis'
    out_path+='.pdf'
    print('A plot has been written to \''+out_path+'\'.')
    savefig(out_path, bbox_inches='tight', transparent=True)
    close()

def plot_grid(evaluation,out_path,value='iterations',name_color=name_color,param_style=param_style):
    fig = figure()

    label_vals = averaged_function_of(evaluation,variable='n_z',value=value,name_color=name_color,param_style=param_style)

    #xlim([min(self.reduced_data[key]['grid'])-5,max(self.reduced_data[key]['grid'])+5])
    ylim([10,30000])

    x_min=1e10; x_max=-1e10;
    for line in gca().lines:
        #import pdb; pdb.set_trace()
        lin=line.get_xdata()
        if len(lin)<1: break
        x_min=min(x_min,lin[0])
        x_max=max(x_max,lin[-1])
    dist=(x_max-x_min)/48.;
    xlim([x_min-dist,x_max+dist])

    yscale('log')
    xlabel('Number of grid points in each direction')
    ylabel(variable_name[value])

    leg = legend(label_vals,loc='best', prop={'size':10},markerscale=0.)
    leg.get_frame().set_alpha(0.5)
    draw()

    out_path+='_nz';
    out_path+='.pdf';
    print('A plot has been written to \''+out_path+'\'.')
    savefig(out_path, bbox_inches='tight', transparent=True)
    close()

def data_plotter(data,pos=0,path="plots",vmin=None,vmax=None):
    element=data.data[pos]
    if not os.path.exists(path): os.makedirs(path)
    outpath=path+"/n_{1}-xi_{0}-{2}-{3}-{4}.pdf".format(element['xi'],element['n_z'],element['solver'],element['boundary'],element['parameters'])
    if os.path.exists(outpath): return
    #import pdb; pdb.set_trace()
    if "data" not in element:
        print('data_plotter: No grid data found in element at position',pos,'.')
        return
    grid,mu,z=data.get_grid(element);
    grid=grid[:,::-1]

    fig=figure()
    ## figure configuration
    n_levels = 5
    n_offset = 0
    if vmin == None: vmin=min(grid.flatten())
    if vmax == None: vmax=max(grid.flatten())
    c_min = 0.88*vmin+0.12*vmax
    c_max = 0.1*vmin+0.9*vmax
    lvls = c_min + arange(n_offset,n_levels+n_offset) * (c_max-c_min)/(n_levels+2*n_offset+1)

    ax = gca()
    im = imshow(grid,extent=(min(z),max(z),max(mu),min(mu)),vmin=vmin,vmax=vmax,aspect='.5',alpha=0.75);
    cntr = contour(grid[::-1,:],extent=(min(z),max(z),max(mu),min(mu)),colors=['k'],levels=lvls,use_clabeltext=True,antialiased=True)

    xlabel('$\zeta$')
    ylabel('$\mu$')

    divider = make_axes_locatable(ax)
    cax = divider.append_axes("top", size="2%", pad=-1.45)
    #caxes = data['plot'].
    #cax, kw = make_axes(data['plot'],location='top',orientation='horizontal')

    cbar = colorbar(im,orientation='horizontal', cax=cax)
    cbar.ax.tick_params(labelsize=10,labeltop=True,labelbottom=False)
    #cbar.set_label_position('top')

    tick_locator = ticker.MaxNLocator(nbins=6)
    cbar.locator = tick_locator
    cbar.update_ticks()


    ## end of figure configuration

    try:
        fig.savefig(outpath, bbox_inches='tight', transparent=True)
    except:
        print('Failed to save plot \''+outpath+'\'.')
    else:
        print('Saved plot\''+outpath+'\'.')

    close()

def fitting():
    pass

if __name__ == "__main__":
    store=data();
    value='iterations'
    out_name='simulation_plots'
    boundaries=None;
    solvers=None;

    ## Import and concatenate all supplied json files.
    for val in sys.argv[1:]:
        if os.path.splitext(val)[1] == ".json":
            store.import_json(val)
            #continue
        else:
            print('The parameter '+val+' has no use.')

    eval=evaluate(store);

    ## generate plots of the reference solutions
    eval.reduce(flags=['reference'])
    for i in range(len(eval.data.data)):
        data_plotter(eval.data,pos=i,path=out_name)
    print('Finished generating grid plots.---------')

    ## Generate plots of the value as a function of xi.
    eval.reduce(boundaries=boundaries,solvers=solvers,xis=None)
    plot_xis(eval,out_path=out_name,value=value)

    ## Generate plots of the value as a function of the grid size.
    eval.reduce(boundaries=boundaries,solvers=solvers,xis=[0.25, 0.5, 0.75, 1.0, 1.25, 1.5, 2.0, 3.0, 4.0, 5.0, 6.0])
    plot_grid(eval,out_path=out_name,value=value)

    eval.reduce(boundaries=["rect"],solvers=['GMRES','GMRES_right_prec_matrixfree'],xis=None)

    plot_grid(eval,out_path=out_name+"_ahpc",value=value,name_color={},param_style={'GMRES': '-', 'GMRES_right_mu_prec': '--', 'Richardson': '-.', 'Richardson_right_prec_matrixfree': '-.', 'GMRES_right_prec_matrixfree': ':'})
