import pylab

from basic import *
from math import isnan
from numpy.linalg import solve
from scipy.sparse.linalg import gmres
from sys import exit
import datetime

class solver(object):
    """This is the base class to all functions that aim to solve Ax=b for x efficiently. The base class includes the reference solver."""
    def __init__(self,parameter,tol=1e-5,max_iter=3e5):
        """The size of the problem is directly taken from the matrix."""
        self.name=str(self.__class__.__name__)
        self.parameters = parameter;
        self.n = self.parameters.n_z*self.parameters.n_mu

        """The tolerance level at which the algorithm terminates."""
        self.tolerance = tol;

        """The maximum number of iterations allowed."""
        self.max_iter = max_iter;
        ## Catch diverging algorithms. The values have to be high enough in order to avoid false positives.
        self.max_res=1e10
        self.count_res_growth=0;
        self.max_count_res_growth=10
        self.res_old=1e9

        """The Results """
        self.state=0; ## 0 uncomputed, +1 successful computation, -1 unsuccessful/diverging computation
        self.iterations = 0;
        self.result = zeros(self.n);

        self.time=0.;
        self.error=None;

    def start(self):
        if self.state==1:
            print('Error: this solver seems to have run already. Stopping execution!')
            exit(1);
        if self.iterations != 0:
            print('Error: something went wrong and the iterations count is non-zero! Exiting!')
            exit(2)
        self.time_start = t_start = datetime.datetime.now();

    def stop(self,state=0):
        if not hasattr(self,'time_start'):
            print('Error: some timing mismatch occurred. Stopping execution!')
            exit(3);
        self.time = (datetime.datetime.now() - self.time_start).total_seconds();
        self.state=state

    def gmres_state_translation(self,state):
        """Translates the output state of the scipy gmres function to the states in this class."""
        if state==0: return 1
        if state>0: return -1
        if state<0: return -2

    def callback(self,res=None):
        """A simple function that is called by the specific solvers after every iterations. It is used to count the number of iterations. The numpy GMRES solver hands over the individual residual."""
        self.iterations+=1;
        if self.iterations>=self.max_iter: return -1

        #print('\n',self.iterations,res)
        ## The next lines aim to catch diverging algorithms. This is only used with the Richardson iteration.
        if isnan(res): return -3
        if res<self.tolerance: return 1
        if res> self.max_res: return -4
        if res> self.res_old: self.count_res_growth+=1
        if self.count_res_growth>self.max_count_res_growth: return -2
        self.res_old=res
        return 0

    def compare(self,reference):
        """Take a second solution as reference and compute the error accordingly."""
        if self.state==1 and reference.state==1:
            self.error = norm2(self.result-reference.result)
            #print(self.error)

    def solve(self):
        """The reference solver."""
        self.name="reference"
        self.start()
        self.result=solve(self.parameters.matrix.todense(),self.parameters.rhs);
        self.iterations=1;
        self.stop(1);

    def get_feedback(self):
        """Return a string with information about the solver and the results. This is intended to be written to stdout."""
        out=""
        out+=str(self.name)+', '
        out+=str(self.parameters.initial_method)+', '
        out+=str(self.parameters.__class__.__name__)+', '
        out+='n='+str(self.parameters.n_z)+', '
        out+='xi='+str(self.parameters.xi)+', '
        out+='state: '+str(self.state)
        if self.state != 0:
            out+=' - '
            out+='t: '+str(self.time)+'s, '
            out+='i: '+str(self.iterations)+', '
            out+='e: '+str(self.error)+', '
            out+='\n'
        return out

class GMRES(solver):
    """Default unpreconditioned GMRES solver. (no)"""
    def solve(self):
        self.start();
        self.result,state=gmres(self.parameters.matrix,self.parameters.rhs,x0=self.parameters.u0,callback=self.callback,tol=self.tolerance,maxiter=self.max_iter)
        self.stop(self.gmres_state_translation(state));

class GMRES_julia(solver):
    """Default unpreconditioned GMRES solver. (no)"""
    def solve(self):
        try:
            import julia
        except:
            print("Could not import \"julia\".")
            exit(1)
        print('test')
        jul = julia.Julia()
        jul.using('IterativeSolvers')
        self.start();
        self.result,state=jul.gmres(self.parameters.u0,self.parameters.matrix,self.parameters.rhs,log=True,tol=self.tolerance,maxiter=self.max_iter)
        import pdb; pdb.set_trace()
        self.stop(self.gmres_state_translation(state));

class GMRES_left_mu_prec(solver):
    """Left preconditioned GMRES solver."""
    def solve(self):
        M=self.parameters.get_Ainv()
        self.start();
        self.result,state=gmres(self.parameters.matrix,self.parameters.rhs,x0=self.parameters.u0,callback=self.callback,tol=self.tolerance,maxiter=self.max_iter,M=M)
        self.stop(self.gmres_state_translation(state));

class GMRES_right_mu_prec(solver):
    """Right preconditioned GMRES solver. (Ainv)"""
    def solve(self):
        M=self.parameters.get_Ainv();
        self.start();
        self.result,state=gmres(self.parameters.matrix.dot(M),self.parameters.rhs,x0=self.parameters.u0,callback=self.callback,tol=self.tolerance,maxiter=self.max_iter)
        self.result=M.dot(self.result);
        self.stop(self.gmres_state_translation(state));

class GMRES_right_prec_matrixfree(solver):
    """Matrix free right preconditioned GMRES solver. (lADI)"""
    def solve(self):
        M=self.parameters.get_linop()
        def linop(x):
            return self.parameters.matrix.dot(M.matvec(x))
        A = sparse.linalg.LinearOperator((self.n,self.n),linop,dtype=float)
        #import pdb; pdb.set_trace()
        self.start()
        self.result,state=gmres(A,self.parameters.rhs,x0=self.parameters.u0,callback=self.callback,tol=self.tolerance,maxiter=self.max_iter)
        self.result=M.matvec(self.result)
        self.stop(self.gmres_state_translation(state));

class Richardson(solver):
    """Preconditioned Richardson iteration. (rADI)"""
    def solve(self,omega=1e-1):
        self.start()
        #import pdb;
        state=0;
        while True:
            #pdb.set_trace()
            update=self.parameters.rhs - self.parameters.matrix.dot(u_ADIr)
            res=norm2(update)
            u_ADIr -= omega*update
            state=self.callback(res)
            if state != 1: break
        self.stop(state)
        self.result = u_ADIr

class Richardson_right_prec_matrixfree(solver):
    """Preconditioned Richardson iteration. (rADI)"""
    def solve(self,omega=1e-1):
        u_ADIr=self.parameters.u0.copy()#.reshape((self.n,1))
        #rhs=to_column(rhs,p).reshape((self.n,1))
        ADIinv=self.parameters.get_linop();

        state=0
        self.start()
        #import pdb; pdb.set_trace()
        while True:
            update=self.parameters.rhs - self.parameters.matrix.dot(ADIinv.matvec(u_ADIr))
            res=norm2(update)
            state=self.callback(res)
            if state != 0: break
            u_ADIr -= omega*update
        self.stop(state)
        self.result = ADIinv.dot(u_ADIr)
